defmodule Clickhouse.Format do
  @moduledoc """
  A behaviour to implement data formatting of Clickhouse results.
  """

  @type t :: module()

  @callback decode(Clickhouse.Result.t()) ::
              {:ok, Clickhouse.Result.t()} | {:error, Clickhouse.Error.t()}

  alias Clickhouse.Error

  @formats %{
    Clickhouse.Format.TSV => [
      "TabSeparated",
      "TSV"
    ]
  }

  @spec find(binary) :: {:ok, Clickhouse.Format.t()} | {:error, Clickhouse.Error.t()}
  for {module, names} <- @formats do
    for name <- names do
      def find(unquote(name)), do: {:ok, unquote(module)}
    end
  end

  def find(format) do
    {:error, Error.new(:not_implemented, "Format #{format} not implemented.")}
  end

  @spec fallback :: Clickhouse.Format.Fallback
  def fallback, do: Clickhouse.Format.Fallback

  @spec decode(Clickhouse.Result.t()) ::
          {:ok, Clickhouse.Result.t()} | {:error, Clickhouse.Error.t()}
  def decode(result) do
    result.__format__.decode(result)
  rescue
    UndefinedFunctionError ->
      {:error, Error.new(:not_implemented, "Format #{result.__format__} not implemented.")}
  end
end
